"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AppType = exports.Env = exports.SyncStartupOperation = exports.SettingSectionSource = exports.SettingStorage = exports.SettingItemSubType = exports.SettingItemType = void 0;
const shim_1 = require("../shim");
const locale_1 = require("../locale");
const eventManager_1 = require("../eventManager");
const BaseModel_1 = require("../BaseModel");
const database_1 = require("../database");
const SyncTargetRegistry_1 = require("../SyncTargetRegistry");
const time_1 = require("../time");
const FileHandler_1 = require("./settings/FileHandler");
const Logger_1 = require("@joplin/utils/Logger");
const mergeGlobalAndLocalSettings_1 = require("../services/profileConfig/mergeGlobalAndLocalSettings");
const splitGlobalAndLocalSettings_1 = require("../services/profileConfig/splitGlobalAndLocalSettings");
const JoplinError_1 = require("../JoplinError");
const { sprintf } = require('sprintf-js');
const ObjectUtils = require('../ObjectUtils');
const { toTitleCase } = require('../string-utils.js');
const { rtrimSlashes, toSystemSlashes } = require('../path-utils');
const logger = Logger_1.default.create('models/Setting');
var SettingItemType;
(function (SettingItemType) {
    SettingItemType[SettingItemType["Int"] = 1] = "Int";
    SettingItemType[SettingItemType["String"] = 2] = "String";
    SettingItemType[SettingItemType["Bool"] = 3] = "Bool";
    SettingItemType[SettingItemType["Array"] = 4] = "Array";
    SettingItemType[SettingItemType["Object"] = 5] = "Object";
    SettingItemType[SettingItemType["Button"] = 6] = "Button";
})(SettingItemType || (exports.SettingItemType = SettingItemType = {}));
var SettingItemSubType;
(function (SettingItemSubType) {
    SettingItemSubType["FilePathAndArgs"] = "file_path_and_args";
    SettingItemSubType["FilePath"] = "file_path";
    SettingItemSubType["DirectoryPath"] = "directory_path";
})(SettingItemSubType || (exports.SettingItemSubType = SettingItemSubType = {}));
var SettingStorage;
(function (SettingStorage) {
    SettingStorage[SettingStorage["Database"] = 1] = "Database";
    SettingStorage[SettingStorage["File"] = 2] = "File";
})(SettingStorage || (exports.SettingStorage = SettingStorage = {}));
var SettingSectionSource;
(function (SettingSectionSource) {
    SettingSectionSource[SettingSectionSource["Default"] = 1] = "Default";
    SettingSectionSource[SettingSectionSource["Plugin"] = 2] = "Plugin";
})(SettingSectionSource || (exports.SettingSectionSource = SettingSectionSource = {}));
var SyncStartupOperation;
(function (SyncStartupOperation) {
    SyncStartupOperation[SyncStartupOperation["None"] = 0] = "None";
    SyncStartupOperation[SyncStartupOperation["ClearLocalSyncState"] = 1] = "ClearLocalSyncState";
    SyncStartupOperation[SyncStartupOperation["ClearLocalData"] = 2] = "ClearLocalData";
})(SyncStartupOperation || (exports.SyncStartupOperation = SyncStartupOperation = {}));
var Env;
(function (Env) {
    Env["Undefined"] = "SET_ME";
    Env["Dev"] = "dev";
    Env["Prod"] = "prod";
})(Env || (exports.Env = Env = {}));
var AppType;
(function (AppType) {
    AppType["Desktop"] = "desktop";
    AppType["Mobile"] = "mobile";
    AppType["Cli"] = "cli";
})(AppType || (exports.AppType = AppType = {}));
// To create a default migration:
//
// - Set the new default value in the setting metadata
// - Add an entry below with the name of the setting and the **previous**
//   default value.
//
// **Never** removes an item from this array, as the array index is essentially
// the migration ID.
const defaultMigrations = [
    {
        name: 'sync.target',
        previousDefault: 7,
    },
    {
        name: 'style.editor.contentMaxWidth',
        previousDefault: 600,
    },
];
const userSettingMigration = [
    {
        oldName: 'spellChecker.language',
        newName: 'spellChecker.languages',
        transformValue: (value) => { return [value]; },
    },
];
class Setting extends BaseModel_1.default {
    static tableName() {
        return 'settings';
    }
    static modelType() {
        return BaseModel_1.default.TYPE_SETTING;
    }
    static reset() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.saveTimeoutId_)
                shim_1.default.clearTimeout(this.saveTimeoutId_);
            if (this.changeEventTimeoutId_)
                shim_1.default.clearTimeout(this.changeEventTimeoutId_);
            this.saveTimeoutId_ = null;
            this.changeEventTimeoutId_ = null;
            this.metadata_ = null;
            this.keys_ = null;
            this.cache_ = [];
            this.customMetadata_ = {};
            this.fileHandler_ = null;
            this.rootFileHandler_ = null;
        });
    }
    static get settingFilePath() {
        return `${this.value('profileDir')}/${this.settingFilename_}`;
    }
    static get rootSettingFilePath() {
        return `${this.value('rootProfileDir')}/${this.settingFilename_}`;
    }
    static get settingFilename() {
        return this.settingFilename_;
    }
    static set settingFilename(v) {
        this.settingFilename_ = v;
    }
    static get fileHandler() {
        if (!this.fileHandler_) {
            this.fileHandler_ = new FileHandler_1.default(this.settingFilePath);
        }
        return this.fileHandler_;
    }
    static get rootFileHandler() {
        if (!this.rootFileHandler_) {
            this.rootFileHandler_ = new FileHandler_1.default(this.rootSettingFilePath);
        }
        return this.rootFileHandler_;
    }
    static keychainService() {
        if (!this.keychainService_)
            throw new Error('keychainService has not been set!!');
        return this.keychainService_;
    }
    static setKeychainService(s) {
        this.keychainService_ = s;
    }
    static metadata() {
        if (this.metadata_)
            return this.metadata_;
        const platform = shim_1.default.platformName();
        const mobilePlatform = shim_1.default.mobilePlatform();
        let wysiwygYes = '';
        let wysiwygNo = '';
        if (shim_1.default.isElectron()) {
            wysiwygYes = ` ${(0, locale_1._)('(wysiwyg: %s)', (0, locale_1._)('yes'))}`;
            wysiwygNo = ` ${(0, locale_1._)('(wysiwyg: %s)', (0, locale_1._)('no'))}`;
        }
        const emptyDirWarning = (0, locale_1._)('Attention: If you change this location, make sure you copy all your content to it before syncing, otherwise all files will be removed! See the FAQ for more details: %s', 'https://joplinapp.org/help/faq');
        // A "public" setting means that it will show up in the various config screens (or config command for the CLI tool), however
        // if if private a setting might still be handled and modified by the app. For instance, the settings related to sorting notes are not
        // public for the mobile and desktop apps because they are handled separately in menus.
        const themeOptions = () => {
            const output = {};
            output[Setting.THEME_LIGHT] = (0, locale_1._)('Light');
            output[Setting.THEME_DARK] = (0, locale_1._)('Dark');
            output[Setting.THEME_DRACULA] = (0, locale_1._)('Dracula');
            output[Setting.THEME_SOLARIZED_LIGHT] = (0, locale_1._)('Solarised Light');
            output[Setting.THEME_SOLARIZED_DARK] = (0, locale_1._)('Solarised Dark');
            output[Setting.THEME_NORD] = (0, locale_1._)('Nord');
            output[Setting.THEME_ARITIM_DARK] = (0, locale_1._)('Aritim Dark');
            output[Setting.THEME_OLED_DARK] = (0, locale_1._)('OLED Dark');
            return output;
        };
        this.buildInMetadata_ = {
            'clientId': {
                value: '',
                type: SettingItemType.String,
                public: false,
            },
            'editor.codeView': {
                value: true,
                type: SettingItemType.Bool,
                public: false,
                appTypes: [AppType.Desktop],
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'sync.openSyncWizard': {
                value: null,
                type: SettingItemType.Button,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Open Sync Wizard...'),
                hideLabel: true,
                section: 'sync',
            },
            'sync.target': {
                value: 0,
                type: SettingItemType.Int,
                isEnum: true,
                public: true,
                section: 'sync',
                label: () => (0, locale_1._)('Synchronisation target'),
                description: (appType) => {
                    return appType !== 'cli' ? null : (0, locale_1._)('The target to synchronise to. Each sync target may have additional parameters which are named as `sync.NUM.NAME` (all documented below).');
                },
                options: () => {
                    return SyncTargetRegistry_1.default.idAndLabelPlainObject(platform);
                },
                optionsOrder: () => {
                    return SyncTargetRegistry_1.default.optionsOrder();
                },
                storage: SettingStorage.File,
            },
            'sync.upgradeState': {
                value: Setting.SYNC_UPGRADE_STATE_IDLE,
                type: SettingItemType.Int,
                public: false,
            },
            'sync.startupOperation': {
                value: SyncStartupOperation.None,
                type: SettingItemType.Int,
                public: false,
            },
            'sync.2.path': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    try {
                        return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('filesystem');
                    }
                    catch (error) {
                        return false;
                    }
                },
                filter: (value) => {
                    return value ? rtrimSlashes(value) : '';
                },
                public: true,
                label: () => (0, locale_1._)('Directory to synchronise with (absolute path)'),
                description: () => emptyDirWarning,
                storage: SettingStorage.File,
            },
            'sync.5.path': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('nextcloud');
                },
                public: true,
                label: () => (0, locale_1._)('Nextcloud WebDAV URL'),
                description: () => emptyDirWarning,
                storage: SettingStorage.File,
            },
            'sync.5.username': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('nextcloud');
                },
                public: true,
                label: () => (0, locale_1._)('Nextcloud username'),
                storage: SettingStorage.File,
            },
            'sync.5.password': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('nextcloud');
                },
                public: true,
                label: () => (0, locale_1._)('Nextcloud password'),
                secure: true,
            },
            'sync.6.path': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('webdav');
                },
                public: true,
                label: () => (0, locale_1._)('WebDAV URL'),
                description: () => emptyDirWarning,
                storage: SettingStorage.File,
            },
            'sync.6.username': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('webdav');
                },
                public: true,
                label: () => (0, locale_1._)('WebDAV username'),
                storage: SettingStorage.File,
            },
            'sync.6.password': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('webdav');
                },
                public: true,
                label: () => (0, locale_1._)('WebDAV password'),
                secure: true,
            },
            'sync.8.path': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    try {
                        return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                    }
                    catch (error) {
                        return false;
                    }
                },
                filter: value => {
                    return value ? rtrimSlashes(value) : '';
                },
                public: true,
                label: () => (0, locale_1._)('S3 bucket'),
                description: () => emptyDirWarning,
                storage: SettingStorage.File,
            },
            'sync.8.url': {
                value: 'https://s3.amazonaws.com/',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                },
                filter: value => {
                    return value ? value.trim() : '';
                },
                public: true,
                label: () => (0, locale_1._)('S3 URL'),
                storage: SettingStorage.File,
            },
            'sync.8.region': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                },
                filter: value => {
                    return value ? value.trim() : '';
                },
                public: true,
                label: () => (0, locale_1._)('S3 region'),
                storage: SettingStorage.File,
            },
            'sync.8.username': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                },
                public: true,
                label: () => (0, locale_1._)('S3 access key'),
                storage: SettingStorage.File,
            },
            'sync.8.password': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                },
                public: true,
                label: () => (0, locale_1._)('S3 secret key'),
                secure: true,
            },
            'sync.8.forcePathStyle': {
                value: false,
                type: SettingItemType.Bool,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('amazon_s3');
                },
                public: true,
                label: () => (0, locale_1._)('Force path style'),
                storage: SettingStorage.File,
            },
            'sync.9.path': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('joplinServer');
                },
                public: true,
                label: () => (0, locale_1._)('Joplin Server URL'),
                description: () => emptyDirWarning,
                storage: SettingStorage.File,
            },
            'sync.9.userContentPath': {
                value: '',
                type: SettingItemType.String,
                public: false,
                storage: SettingStorage.Database,
            },
            'sync.9.username': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('joplinServer');
                },
                public: true,
                label: () => (0, locale_1._)('Joplin Server email'),
                storage: SettingStorage.File,
            },
            'sync.9.password': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('joplinServer');
                },
                public: true,
                label: () => (0, locale_1._)('Joplin Server password'),
                secure: true,
            },
            // Although sync.10.path is essentially a constant, we still define
            // it here so that both Joplin Server and Joplin Cloud can be
            // handled in the same consistent way. Also having it a setting
            // means it can be set to something else for development.
            'sync.10.path': {
                value: 'https://api.joplincloud.com',
                type: SettingItemType.String,
                public: false,
                storage: SettingStorage.Database,
            },
            'sync.10.userContentPath': {
                value: 'https://joplinusercontent.com',
                type: SettingItemType.String,
                public: false,
                storage: SettingStorage.Database,
            },
            'sync.10.username': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('joplinCloud');
                },
                public: true,
                label: () => (0, locale_1._)('Joplin Cloud email'),
                storage: SettingStorage.File,
            },
            'sync.10.password': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                show: (settings) => {
                    return settings['sync.target'] === SyncTargetRegistry_1.default.nameToId('joplinCloud');
                },
                public: true,
                label: () => (0, locale_1._)('Joplin Cloud password'),
                secure: true,
            },
            'sync.10.inboxEmail': { value: '', type: SettingItemType.String, public: false },
            'sync.10.inboxId': { value: '', type: SettingItemType.String, public: false },
            'sync.10.canUseSharePermissions': { value: false, type: SettingItemType.Bool, public: false },
            'sync.10.accountType': { value: 0, type: SettingItemType.Int, public: false },
            'sync.5.syncTargets': { value: {}, type: SettingItemType.Object, public: false },
            'sync.resourceDownloadMode': {
                value: 'always',
                type: SettingItemType.String,
                section: 'sync',
                public: true,
                advanced: true,
                isEnum: true,
                appTypes: [AppType.Mobile, AppType.Desktop],
                label: () => (0, locale_1._)('Attachment download behaviour'),
                description: () => (0, locale_1._)('In "Manual" mode, attachments are downloaded only when you click on them. In "Auto", they are downloaded when you open the note. In "Always", all the attachments are downloaded whether you open the note or not.'),
                options: () => {
                    return {
                        always: (0, locale_1._)('Always'),
                        manual: (0, locale_1._)('Manual'),
                        auto: (0, locale_1._)('Auto'),
                    };
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'sync.3.auth': { value: '', type: SettingItemType.String, public: false },
            'sync.4.auth': { value: '', type: SettingItemType.String, public: false },
            'sync.7.auth': { value: '', type: SettingItemType.String, public: false },
            'sync.9.auth': { value: '', type: SettingItemType.String, public: false },
            'sync.10.auth': { value: '', type: SettingItemType.String, public: false },
            'sync.1.context': { value: '', type: SettingItemType.String, public: false },
            'sync.2.context': { value: '', type: SettingItemType.String, public: false },
            'sync.3.context': { value: '', type: SettingItemType.String, public: false },
            'sync.4.context': { value: '', type: SettingItemType.String, public: false },
            'sync.5.context': { value: '', type: SettingItemType.String, public: false },
            'sync.6.context': { value: '', type: SettingItemType.String, public: false },
            'sync.7.context': { value: '', type: SettingItemType.String, public: false },
            'sync.8.context': { value: '', type: SettingItemType.String, public: false },
            'sync.9.context': { value: '', type: SettingItemType.String, public: false },
            'sync.10.context': { value: '', type: SettingItemType.String, public: false },
            'sync.maxConcurrentConnections': { value: 5, type: SettingItemType.Int, storage: SettingStorage.File, isGlobal: true, public: true, advanced: true, section: 'sync', label: () => (0, locale_1._)('Max concurrent connections'), minimum: 1, maximum: 20, step: 1 },
            // The active folder ID is guaranteed to be valid as long as there's at least one
            // existing folder, so it is a good default in contexts where there's no currently
            // selected folder. It corresponds in general to the currently selected folder or
            // to the last folder that was selected.
            activeFolderId: { value: '', type: SettingItemType.String, public: false },
            notesParent: { value: '', type: SettingItemType.String, public: false },
            richTextBannerDismissed: { value: false, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, public: false },
            firstStart: { value: true, type: SettingItemType.Bool, public: false },
            locale: {
                value: (0, locale_1.defaultLocale)(),
                type: SettingItemType.String,
                isEnum: true,
                public: true,
                label: () => (0, locale_1._)('Language'),
                options: () => {
                    return ObjectUtils.sortByValue((0, locale_1.supportedLocalesToLanguages)({ includeStats: true }));
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            dateFormat: {
                value: Setting.DATE_FORMAT_1,
                type: SettingItemType.String,
                isEnum: true,
                public: true,
                label: () => (0, locale_1._)('Date format'),
                options: () => {
                    const options = {};
                    const now = new Date('2017-01-30T12:00:00').getTime();
                    options[Setting.DATE_FORMAT_1] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_1);
                    options[Setting.DATE_FORMAT_2] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_2);
                    options[Setting.DATE_FORMAT_3] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_3);
                    options[Setting.DATE_FORMAT_4] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_4);
                    options[Setting.DATE_FORMAT_5] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_5);
                    options[Setting.DATE_FORMAT_6] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_6);
                    options[Setting.DATE_FORMAT_7] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_7);
                    options[Setting.DATE_FORMAT_8] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_8);
                    options[Setting.DATE_FORMAT_9] = time_1.default.formatMsToLocal(now, Setting.DATE_FORMAT_9);
                    return options;
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            timeFormat: {
                value: Setting.TIME_FORMAT_1,
                type: SettingItemType.String,
                isEnum: true,
                public: true,
                label: () => (0, locale_1._)('Time format'),
                options: () => {
                    const options = {};
                    const now = new Date('2017-01-30T20:30:00').getTime();
                    options[Setting.TIME_FORMAT_1] = time_1.default.formatMsToLocal(now, Setting.TIME_FORMAT_1);
                    options[Setting.TIME_FORMAT_2] = time_1.default.formatMsToLocal(now, Setting.TIME_FORMAT_2);
                    options[Setting.TIME_FORMAT_3] = time_1.default.formatMsToLocal(now, Setting.TIME_FORMAT_3);
                    return options;
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'ocr.enabled': {
                value: false,
                type: SettingItemType.Bool,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Enable optical character recognition (OCR)'),
                description: () => (0, locale_1._)('When enabled, the application will scan your attachments and extract the text from it. This will allow you to search for text in these attachments.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            theme: {
                value: Setting.THEME_LIGHT,
                type: SettingItemType.Int,
                public: true,
                appTypes: [AppType.Mobile, AppType.Desktop],
                show: (settings) => {
                    return !settings['themeAutoDetect'];
                },
                isEnum: true,
                label: () => (0, locale_1._)('Theme'),
                section: 'appearance',
                options: () => themeOptions(),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            themeAutoDetect: {
                value: false,
                type: SettingItemType.Bool,
                section: 'appearance',
                appTypes: [AppType.Mobile, AppType.Desktop],
                public: true,
                label: () => (0, locale_1._)('Automatically switch theme to match system theme'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            preferredLightTheme: {
                value: Setting.THEME_LIGHT,
                type: SettingItemType.Int,
                public: true,
                show: (settings) => {
                    return settings['themeAutoDetect'];
                },
                appTypes: [AppType.Mobile, AppType.Desktop],
                isEnum: true,
                label: () => (0, locale_1._)('Preferred light theme'),
                section: 'appearance',
                options: () => themeOptions(),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            preferredDarkTheme: {
                value: Setting.THEME_DARK,
                type: SettingItemType.Int,
                public: true,
                show: (settings) => {
                    return settings['themeAutoDetect'];
                },
                appTypes: [AppType.Mobile, AppType.Desktop],
                isEnum: true,
                label: () => (0, locale_1._)('Preferred dark theme'),
                section: 'appearance',
                options: () => themeOptions(),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            notificationPermission: {
                value: '',
                type: SettingItemType.String,
                public: false,
            },
            showNoteCounts: { value: true, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, public: false, advanced: true, appTypes: [AppType.Desktop, AppType.Cli], label: () => (0, locale_1._)('Show note counts') },
            layoutButtonSequence: {
                value: Setting.LAYOUT_ALL,
                type: SettingItemType.Int,
                public: false,
                appTypes: [AppType.Desktop],
                isEnum: true,
                options: () => ({
                    [Setting.LAYOUT_ALL]: (0, locale_1._)('%s / %s / %s', (0, locale_1._)('Editor'), (0, locale_1._)('Viewer'), (0, locale_1._)('Split View')),
                    [Setting.LAYOUT_EDITOR_VIEWER]: (0, locale_1._)('%s / %s', (0, locale_1._)('Editor'), (0, locale_1._)('Viewer')),
                    [Setting.LAYOUT_EDITOR_SPLIT]: (0, locale_1._)('%s / %s', (0, locale_1._)('Editor'), (0, locale_1._)('Split View')),
                    [Setting.LAYOUT_VIEWER_SPLIT]: (0, locale_1._)('%s / %s', (0, locale_1._)('Viewer'), (0, locale_1._)('Split View')),
                }),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            uncompletedTodosOnTop: { value: true, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, section: 'note', public: true, appTypes: [AppType.Cli], label: () => (0, locale_1._)('Uncompleted to-dos on top') },
            showCompletedTodos: { value: true, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, section: 'note', public: true, appTypes: [AppType.Cli], label: () => (0, locale_1._)('Show completed to-dos') },
            'notes.sortOrder.field': {
                value: 'user_updated_time',
                type: SettingItemType.String,
                section: 'note',
                isEnum: true,
                public: true,
                appTypes: [AppType.Cli],
                label: () => (0, locale_1._)('Sort notes by'),
                options: () => {
                    const Note = require('./Note').default;
                    const noteSortFields = ['user_updated_time', 'user_created_time', 'title', 'order'];
                    const options = {};
                    for (let i = 0; i < noteSortFields.length; i++) {
                        options[noteSortFields[i]] = toTitleCase(Note.fieldToLabel(noteSortFields[i]));
                    }
                    return options;
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'editor.autoMatchingBraces': {
                value: true,
                type: SettingItemType.Bool,
                public: true,
                section: 'note',
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Auto-pair braces, parenthesis, quotations, etc.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'notes.sortOrder.reverse': { value: true, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, section: 'note', public: true, label: () => (0, locale_1._)('Reverse sort order'), appTypes: [AppType.Cli] },
            // NOTE: A setting whose name starts with 'notes.sortOrder' is special,
            // which implies changing the setting automatically triggers the refresh of notes.
            // See lib/BaseApplication.ts/generalMiddleware() for details.
            'notes.sortOrder.buttonsVisible': {
                value: true,
                type: SettingItemType.Bool,
                storage: SettingStorage.File,
                section: 'appearance',
                public: true,
                label: () => (0, locale_1._)('Show sort order buttons'),
                // description: () => _('If true, sort order buttons (field + reverse) for notes are shown at the top of Note List.'),
                appTypes: [AppType.Desktop],
                isGlobal: true,
            },
            'notes.perFieldReversalEnabled': {
                value: true,
                type: SettingItemType.Bool,
                storage: SettingStorage.File,
                section: 'note',
                public: false,
                appTypes: [AppType.Cli, AppType.Desktop],
            },
            'notes.perFieldReverse': {
                value: {
                    user_updated_time: true,
                    user_created_time: true,
                    title: false,
                    order: false,
                },
                type: SettingItemType.Object,
                storage: SettingStorage.File,
                section: 'note',
                public: false,
                appTypes: [AppType.Cli, AppType.Desktop],
            },
            'notes.perFolderSortOrderEnabled': {
                value: true,
                type: SettingItemType.Bool,
                storage: SettingStorage.File,
                section: 'folder',
                public: false,
                appTypes: [AppType.Cli, AppType.Desktop],
            },
            'notes.perFolderSortOrders': {
                value: {},
                type: SettingItemType.Object,
                storage: SettingStorage.File,
                section: 'folder',
                public: false,
                appTypes: [AppType.Cli, AppType.Desktop],
            },
            'notes.sharedSortOrder': {
                value: {},
                type: SettingItemType.Object,
                section: 'folder',
                public: false,
                appTypes: [AppType.Cli, AppType.Desktop],
            },
            'folders.sortOrder.field': {
                value: 'title',
                type: SettingItemType.String,
                isEnum: true,
                public: true,
                appTypes: [AppType.Cli],
                label: () => (0, locale_1._)('Sort notebooks by'),
                options: () => {
                    const Folder = require('./Folder').default;
                    const folderSortFields = ['title', 'last_note_user_updated_time'];
                    const options = {};
                    for (let i = 0; i < folderSortFields.length; i++) {
                        options[folderSortFields[i]] = toTitleCase(Folder.fieldToLabel(folderSortFields[i]));
                    }
                    return options;
                },
                storage: SettingStorage.File,
            },
            'folders.sortOrder.reverse': { value: false, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, public: true, label: () => (0, locale_1._)('Reverse sort order'), appTypes: [AppType.Cli] },
            trackLocation: { value: true, type: SettingItemType.Bool, section: 'note', storage: SettingStorage.File, isGlobal: true, public: true, label: () => (0, locale_1._)('Save geo-location with notes') },
            'editor.usePlainText': {
                value: false,
                type: SettingItemType.Bool,
                section: 'note',
                public: true,
                appTypes: [AppType.Mobile],
                label: () => 'Use the plain text editor',
                description: () => 'The plain text editor has various issues and is no longer supported. If you are having issues with the new editor however you can revert to the old one using this setting.',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            // Enables/disables spellcheck in the mobile markdown beta editor.
            'editor.mobile.spellcheckEnabled': {
                value: true,
                type: SettingItemType.Bool,
                section: 'note',
                public: true,
                appTypes: [AppType.Mobile],
                label: () => (0, locale_1._)('Enable spellcheck in the text editor'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'editor.mobile.toolbarEnabled': {
                value: true,
                type: SettingItemType.Bool,
                section: 'note',
                public: true,
                appTypes: [AppType.Mobile],
                label: () => (0, locale_1._)('Enable the Markdown toolbar'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            // Works around a bug in which additional space is visible beneath the toolbar on some devices.
            // See https://github.com/laurent22/joplin/pull/6823
            'editor.mobile.removeSpaceBelowToolbar': {
                value: false,
                type: SettingItemType.Bool,
                section: 'note',
                public: true,
                appTypes: [AppType.Mobile],
                show: (settings) => settings['editor.mobile.removeSpaceBelowToolbar'],
                label: () => 'Remove extra space below the markdown toolbar',
                description: () => 'Works around bug on some devices where the markdown toolbar does not touch the bottom of the screen.',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            newTodoFocus: {
                value: 'title',
                type: SettingItemType.String,
                section: 'note',
                isEnum: true,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('When creating a new to-do:'),
                options: () => {
                    return {
                        title: (0, locale_1._)('Focus title'),
                        body: (0, locale_1._)('Focus body'),
                    };
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            newNoteFocus: {
                value: 'body',
                type: SettingItemType.String,
                section: 'note',
                isEnum: true,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('When creating a new note:'),
                options: () => {
                    return {
                        title: (0, locale_1._)('Focus title'),
                        body: (0, locale_1._)('Focus body'),
                    };
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            imageResizing: {
                value: 'alwaysAsk',
                type: SettingItemType.String,
                section: 'note',
                isEnum: true,
                public: true,
                appTypes: [AppType.Mobile, AppType.Desktop],
                label: () => (0, locale_1._)('Resize large images:'),
                description: () => (0, locale_1._)('Shrink large images before adding them to notes.'),
                options: () => {
                    return {
                        alwaysAsk: (0, locale_1._)('Always ask'),
                        alwaysResize: (0, locale_1._)('Always resize'),
                        neverResize: (0, locale_1._)('Never resize'),
                    };
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'notes.listRendererId': {
                value: 'compact',
                type: SettingItemType.String,
                public: false,
                appTypes: [AppType.Desktop],
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'plugins.states': {
                value: '',
                type: SettingItemType.Object,
                section: 'plugins',
                public: true,
                appTypes: [AppType.Desktop],
                needRestart: true,
                autoSave: true,
            },
            'plugins.devPluginPaths': {
                value: '',
                type: SettingItemType.String,
                section: 'plugins',
                public: true,
                advanced: true,
                appTypes: [AppType.Desktop],
                label: () => 'Development plugins',
                description: () => 'You may add multiple plugin paths, each separated by a comma. You will need to restart the application for the changes to take effect.',
                storage: SettingStorage.File,
            },
            // Deprecated - use markdown.plugin.*
            'markdown.softbreaks': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, public: false, appTypes: [AppType.Mobile, AppType.Desktop] },
            'markdown.typographer': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, public: false, appTypes: [AppType.Mobile, AppType.Desktop] },
            // Deprecated
            'markdown.plugin.softbreaks': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable soft breaks')}${wysiwygYes}` },
            'markdown.plugin.typographer': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable typographer support')}${wysiwygYes}` },
            'markdown.plugin.linkify': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable Linkify')}${wysiwygYes}` },
            'markdown.plugin.katex': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable math expressions')}${wysiwygYes}` },
            'markdown.plugin.fountain': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable Fountain syntax support')}${wysiwygYes}` },
            'markdown.plugin.mermaid': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable Mermaid diagrams support')}${wysiwygYes}` },
            'markdown.plugin.audioPlayer': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable audio player')}${wysiwygNo}` },
            'markdown.plugin.videoPlayer': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable video player')}${wysiwygNo}` },
            'markdown.plugin.pdfViewer': { storage: SettingStorage.File, isGlobal: true, value: !mobilePlatform, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Desktop], label: () => `${(0, locale_1._)('Enable PDF viewer')}${wysiwygNo}` },
            'markdown.plugin.mark': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable ==mark== syntax')}${wysiwygYes}` },
            'markdown.plugin.footnote': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable footnotes')}${wysiwygNo}` },
            'markdown.plugin.toc': { storage: SettingStorage.File, isGlobal: true, value: true, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable table of contents extension')}${wysiwygNo}` },
            'markdown.plugin.sub': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable ~sub~ syntax')}${wysiwygYes}` },
            'markdown.plugin.sup': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable ^sup^ syntax')}${wysiwygYes}` },
            'markdown.plugin.deflist': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable deflist syntax')}${wysiwygNo}` },
            'markdown.plugin.abbr': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable abbreviation syntax')}${wysiwygNo}` },
            'markdown.plugin.emoji': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable markdown emoji')}${wysiwygNo}` },
            'markdown.plugin.insert': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable ++insert++ syntax')}${wysiwygYes}` },
            'markdown.plugin.multitable': { storage: SettingStorage.File, isGlobal: true, value: false, type: SettingItemType.Bool, section: 'markdownPlugins', public: true, appTypes: [AppType.Mobile, AppType.Desktop], label: () => `${(0, locale_1._)('Enable multimarkdown table extension')}${wysiwygNo}` },
            // Tray icon (called AppIndicator) doesn't work in Ubuntu
            // http://www.webupd8.org/2017/04/fix-appindicator-not-working-for.html
            // Might be fixed in Electron 18.x but no non-beta release yet. So for now
            // by default we disable it on Linux.
            showTrayIcon: {
                value: platform !== 'linux',
                type: SettingItemType.Bool,
                section: 'application',
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Show tray icon'),
                description: () => {
                    return platform === 'linux' ? (0, locale_1._)('Note: Does not work in all desktop environments.') : (0, locale_1._)('This will allow Joplin to run in the background. It is recommended to enable this setting so that your notes are constantly being synchronised, thus reducing the number of conflicts.');
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            startMinimized: { value: false, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, section: 'application', public: true, appTypes: [AppType.Desktop], label: () => (0, locale_1._)('Start application minimised in the tray icon') },
            collapsedFolderIds: { value: [], type: SettingItemType.Array, public: false },
            'keychain.supported': { value: -1, type: SettingItemType.Int, public: false },
            'db.ftsEnabled': { value: -1, type: SettingItemType.Int, public: false },
            'db.fuzzySearchEnabled': { value: -1, type: SettingItemType.Int, public: false },
            'encryption.enabled': { value: false, type: SettingItemType.Bool, public: false },
            'encryption.activeMasterKeyId': { value: '', type: SettingItemType.String, public: false },
            'encryption.passwordCache': { value: {}, type: SettingItemType.Object, public: false, secure: true },
            'encryption.masterPassword': { value: '', type: SettingItemType.String, public: false, secure: true },
            'encryption.shouldReencrypt': {
                value: -1,
                type: SettingItemType.Int,
                public: false,
            },
            'sync.userId': {
                value: '',
                type: SettingItemType.String,
                public: false,
            },
            // Deprecated in favour of windowContentZoomFactor
            'style.zoom': { value: 100, type: SettingItemType.Int, public: false, storage: SettingStorage.File, isGlobal: true, appTypes: [AppType.Desktop], section: 'appearance', label: () => '', minimum: 50, maximum: 500, step: 10 },
            'style.editor.fontSize': {
                value: 15,
                type: SettingItemType.Int,
                public: true,
                storage: SettingStorage.File,
                isGlobal: true,
                appTypes: [AppType.Desktop, AppType.Mobile],
                section: 'appearance',
                label: () => (0, locale_1._)('Editor font size'),
                minimum: 4,
                maximum: 50,
                step: 1,
            },
            'style.editor.fontFamily': (mobilePlatform) ?
                ({
                    value: Setting.FONT_DEFAULT,
                    type: SettingItemType.String,
                    isEnum: true,
                    public: true,
                    label: () => (0, locale_1._)('Editor font'),
                    appTypes: [AppType.Mobile],
                    section: 'appearance',
                    options: () => {
                        // IMPORTANT: The font mapping must match the one in global-styles.js::editorFont()
                        if (mobilePlatform === 'ios') {
                            return {
                                [Setting.FONT_DEFAULT]: (0, locale_1._)('Default'),
                                [Setting.FONT_MENLO]: 'Menlo',
                                [Setting.FONT_COURIER_NEW]: 'Courier New',
                                [Setting.FONT_AVENIR]: 'Avenir',
                            };
                        }
                        return {
                            [Setting.FONT_DEFAULT]: (0, locale_1._)('Default'),
                            [Setting.FONT_MONOSPACE]: 'Monospace',
                        };
                    },
                    storage: SettingStorage.File,
                    isGlobal: true,
                }) : {
                value: '',
                type: SettingItemType.String,
                public: true,
                appTypes: [AppType.Desktop],
                section: 'appearance',
                label: () => (0, locale_1._)('Editor font family'),
                description: () => (0, locale_1._)('Used for most text in the markdown editor. If not found, a generic proportional (variable width) font is used.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'style.editor.monospaceFontFamily': {
                value: '',
                type: SettingItemType.String,
                public: true,
                appTypes: [AppType.Desktop],
                section: 'appearance',
                label: () => (0, locale_1._)('Editor monospace font family'),
                description: () => (0, locale_1._)('Used where a fixed width font is needed to lay out text legibly (e.g. tables, checkboxes, code). If not found, a generic monospace (fixed width) font is used.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'style.editor.contentMaxWidth': { value: 0, type: SettingItemType.Int, public: true, storage: SettingStorage.File, isGlobal: true, appTypes: [AppType.Desktop], section: 'appearance', label: () => (0, locale_1._)('Editor maximum width'), description: () => (0, locale_1._)('Set it to 0 to make it take the complete available space. Recommended width is 600.') },
            'ui.layout': { value: {}, type: SettingItemType.Object, storage: SettingStorage.File, isGlobal: true, public: false, appTypes: [AppType.Desktop] },
            // TODO: Is there a better way to do this? The goal here is to simply have
            // a way to display a link to the customizable stylesheets, not for it to
            // serve as a customizable Setting. But because the Setting page is auto-
            // generated from this list of settings, there wasn't a really elegant way
            // to do that directly in the React markup.
            'style.customCss.renderedMarkdown': {
                value: null,
                onClick: () => {
                    shim_1.default.openOrCreateFile(this.customCssFilePath(Setting.customCssFilenames.RENDERED_MARKDOWN), '/* For styling the rendered Markdown */');
                },
                type: SettingItemType.Button,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Custom stylesheet for rendered Markdown'),
                section: 'appearance',
                advanced: true,
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'style.customCss.joplinApp': {
                value: null,
                onClick: () => {
                    shim_1.default.openOrCreateFile(this.customCssFilePath(Setting.customCssFilenames.JOPLIN_APP), `/* For styling the entire Joplin app (except the rendered Markdown, which is defined in \`${Setting.customCssFilenames.RENDERED_MARKDOWN}\`) */`);
                },
                type: SettingItemType.Button,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Custom stylesheet for Joplin-wide app styles'),
                section: 'appearance',
                advanced: true,
                description: () => 'CSS file support is provided for your convenience, but they are advanced settings, and styles you define may break from one version to the next. If you want to use them, please know that it might require regular development work from you to keep them working. The Joplin team cannot make a commitment to keep the application HTML structure stable.',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'sync.clearLocalSyncStateButton': {
                value: null,
                type: SettingItemType.Button,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Re-upload local data to sync target'),
                section: 'sync',
                advanced: true,
                description: () => 'If the data on the sync target is incorrect or empty, you can use this button to force a re-upload of your data to the sync target. Application will have to be restarted',
            },
            'sync.clearLocalDataButton': {
                value: null,
                type: SettingItemType.Button,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => (0, locale_1._)('Delete local data and re-download from sync target'),
                section: 'sync',
                advanced: true,
                description: () => 'If the data on the sync target is correct but your local data is not, you can use this button to clear your local data and force re-downloading everything from the sync target. As your local data will be deleted first, it is recommended to export your data as JEX first. Application will have to be restarted',
            },
            autoUpdateEnabled: { value: true, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, section: 'application', public: platform !== 'linux', appTypes: [AppType.Desktop], label: () => (0, locale_1._)('Automatically check for updates') },
            'autoUpdate.includePreReleases': { value: false, type: SettingItemType.Bool, section: 'application', storage: SettingStorage.File, isGlobal: true, public: true, appTypes: [AppType.Desktop], label: () => (0, locale_1._)('Get pre-releases when checking for updates'), description: () => (0, locale_1._)('See the pre-release page for more details: %s', 'https://joplinapp.org/help/about/prereleases') },
            'autoUploadCrashDumps': {
                value: false,
                section: 'application',
                type: SettingItemType.Bool,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => 'Automatically upload crash reports',
                description: () => 'If you experience a crash, please enable this option to automatically send crash reports. You will need to restart the application for this change to take effect.',
                isGlobal: true,
                storage: SettingStorage.File,
            },
            'clipperServer.autoStart': { value: false, type: SettingItemType.Bool, storage: SettingStorage.File, isGlobal: true, public: false },
            'sync.interval': {
                value: 300,
                type: SettingItemType.Int,
                section: 'sync',
                isEnum: true,
                public: true,
                label: () => (0, locale_1._)('Synchronisation interval'),
                options: () => {
                    return {
                        0: (0, locale_1._)('Disabled'),
                        300: (0, locale_1._)('%d minutes', 5),
                        600: (0, locale_1._)('%d minutes', 10),
                        1800: (0, locale_1._)('%d minutes', 30),
                        3600: (0, locale_1._)('%d hour', 1),
                        43200: (0, locale_1._)('%d hours', 12),
                        86400: (0, locale_1._)('%d hours', 24),
                    };
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'sync.mobileWifiOnly': {
                value: false,
                type: SettingItemType.Bool,
                section: 'sync',
                public: true,
                label: () => (0, locale_1._)('Synchronise only over WiFi connection'),
                storage: SettingStorage.File,
                appTypes: [AppType.Mobile],
                isGlobal: true,
            },
            noteVisiblePanes: { value: ['editor', 'viewer'], type: SettingItemType.Array, storage: SettingStorage.File, isGlobal: true, public: false, appTypes: [AppType.Desktop] },
            tagHeaderIsExpanded: { value: true, type: SettingItemType.Bool, public: false, appTypes: [AppType.Desktop] },
            folderHeaderIsExpanded: { value: true, type: SettingItemType.Bool, public: false, appTypes: [AppType.Desktop] },
            editor: { value: '', type: SettingItemType.String, subType: 'file_path_and_args', storage: SettingStorage.File, isGlobal: true, public: true, appTypes: [AppType.Cli, AppType.Desktop], label: () => (0, locale_1._)('Text editor command'), description: () => (0, locale_1._)('The editor command (may include arguments) that will be used to open a note. If none is provided it will try to auto-detect the default editor.') },
            'export.pdfPageSize': { value: 'A4', type: SettingItemType.String, advanced: true, storage: SettingStorage.File, isGlobal: true, isEnum: true, public: true, appTypes: [AppType.Desktop], label: () => (0, locale_1._)('Page size for PDF export'), options: () => {
                    return {
                        'A4': (0, locale_1._)('A4'),
                        'Letter': (0, locale_1._)('Letter'),
                        'A3': (0, locale_1._)('A3'),
                        'A5': (0, locale_1._)('A5'),
                        'Tabloid': (0, locale_1._)('Tabloid'),
                        'Legal': (0, locale_1._)('Legal'),
                    };
                } },
            'export.pdfPageOrientation': { value: 'portrait', type: SettingItemType.String, storage: SettingStorage.File, isGlobal: true, advanced: true, isEnum: true, public: true, appTypes: [AppType.Desktop], label: () => (0, locale_1._)('Page orientation for PDF export'), options: () => {
                    return {
                        'portrait': (0, locale_1._)('Portrait'),
                        'landscape': (0, locale_1._)('Landscape'),
                    };
                } },
            useCustomPdfViewer: {
                value: false,
                type: SettingItemType.Bool,
                public: false,
                advanced: true,
                appTypes: [AppType.Desktop],
                label: () => 'Use custom PDF viewer (Beta)',
                description: () => 'The custom PDF viewer remembers the last page that was viewed, however it has some technical issues.',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'editor.keyboardMode': {
                value: '',
                type: SettingItemType.String,
                public: true,
                appTypes: [AppType.Desktop],
                isEnum: true,
                advanced: true,
                label: () => (0, locale_1._)('Keyboard Mode'),
                options: () => {
                    const output = {};
                    output[''] = (0, locale_1._)('Default');
                    output['emacs'] = (0, locale_1._)('Emacs');
                    output['vim'] = (0, locale_1._)('Vim');
                    return output;
                },
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'editor.spellcheckBeta': {
                value: false,
                type: SettingItemType.Bool,
                public: true,
                appTypes: [AppType.Desktop],
                label: () => 'Enable spell checking in Markdown editor? (WARNING BETA feature)',
                description: () => 'Spell checker in the Markdown editor was previously unstable (cursor location was not stable, sometimes edits would not be saved or reflected in the viewer, etc.) however it appears to be more reliable now. If you notice any issue, please report it on GitHub or the Joplin Forum (Help -> Joplin Forum)',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'imageeditor.jsdrawToolbar': {
                value: '',
                type: SettingItemType.String,
                public: false,
                appTypes: [AppType.Mobile],
                label: () => '',
                storage: SettingStorage.File,
            },
            'imageeditor.imageTemplate': {
                value: '{ }',
                type: SettingItemType.String,
                public: false,
                appTypes: [AppType.Mobile],
                label: () => 'Template for the image editor',
                storage: SettingStorage.File,
            },
            // 2023-09-07: This setting is now used to track the desktop beta editor. It
            // was used to track the mobile beta editor previously.
            'editor.beta': {
                value: false,
                type: SettingItemType.Bool,
                section: 'general',
                public: true,
                appTypes: [AppType.Desktop],
                label: () => 'Opt-in to the editor beta',
                description: () => 'This beta adds improved accessibility and plugin API compatibility with the mobile editor. If you find bugs, please report them in the Discourse forum.',
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'net.customCertificates': {
                value: '',
                type: SettingItemType.String,
                section: 'sync',
                advanced: true,
                show: (settings) => {
                    return [
                        SyncTargetRegistry_1.default.nameToId('amazon_s3'),
                        SyncTargetRegistry_1.default.nameToId('nextcloud'),
                        SyncTargetRegistry_1.default.nameToId('webdav'),
                        SyncTargetRegistry_1.default.nameToId('joplinServer'),
                    ].indexOf(settings['sync.target']) >= 0;
                },
                public: true,
                appTypes: [AppType.Desktop, AppType.Cli],
                label: () => (0, locale_1._)('Custom TLS certificates'),
                description: () => (0, locale_1._)('Comma-separated list of paths to directories to load the certificates from, or path to individual cert files. For example: /my/cert_dir, /other/custom.pem. Note that if you make changes to the TLS settings, you must save your changes before clicking on "Check synchronisation configuration".'),
                storage: SettingStorage.File,
            },
            'net.ignoreTlsErrors': {
                value: false,
                type: SettingItemType.Bool,
                advanced: true,
                section: 'sync',
                show: (settings) => {
                    return (shim_1.default.isNode() || shim_1.default.mobilePlatform() === 'android') &&
                        [
                            SyncTargetRegistry_1.default.nameToId('amazon_s3'),
                            SyncTargetRegistry_1.default.nameToId('nextcloud'),
                            SyncTargetRegistry_1.default.nameToId('webdav'),
                            SyncTargetRegistry_1.default.nameToId('joplinServer'),
                            // Needs to be enabled for Joplin Cloud too because
                            // some companies filter all traffic and swap TLS
                            // certificates, which result in error
                            // UNABLE_TO_GET_ISSUER_CERT_LOCALLY
                            SyncTargetRegistry_1.default.nameToId('joplinCloud'),
                        ].indexOf(settings['sync.target']) >= 0;
                },
                public: true,
                label: () => (0, locale_1._)('Ignore TLS certificate errors'),
                storage: SettingStorage.File,
            },
            'net.proxyEnabled': {
                value: false,
                type: SettingItemType.Bool,
                advanced: true,
                section: 'sync',
                isGlobal: true,
                public: true,
                label: () => (0, locale_1._)('Proxy enabled'),
                storage: SettingStorage.File,
            },
            'net.proxyUrl': {
                value: '',
                type: SettingItemType.String,
                advanced: true,
                section: 'sync',
                isGlobal: true,
                public: true,
                label: () => (0, locale_1._)('Proxy URL'),
                description: () => (0, locale_1._)('For example "%s"', 'http://my.proxy.com:80'),
                storage: SettingStorage.File,
            },
            'net.proxyTimeout': {
                value: 1,
                type: SettingItemType.Int,
                maximum: 60,
                advanced: true,
                section: 'sync',
                isGlobal: true,
                public: true,
                label: () => (0, locale_1._)('Proxy timeout (seconds)'),
                storage: SettingStorage.File,
            },
            'sync.wipeOutFailSafe': {
                value: true,
                type: SettingItemType.Bool,
                advanced: true,
                public: true,
                section: 'sync',
                label: () => (0, locale_1._)('Fail-safe'),
                description: () => (0, locale_1._)('Fail-safe: Do not wipe out local data when sync target is empty (often the result of a misconfiguration or bug)'),
                storage: SettingStorage.File,
            },
            'api.token': { value: null, type: SettingItemType.String, public: false, storage: SettingStorage.File, isGlobal: true },
            'api.port': { value: null, type: SettingItemType.Int, storage: SettingStorage.File, isGlobal: true, public: true, appTypes: [AppType.Cli], description: () => (0, locale_1._)('Specify the port that should be used by the API server. If not set, a default will be used.') },
            'resourceService.lastProcessedChangeId': { value: 0, type: SettingItemType.Int, public: false },
            'searchEngine.lastProcessedChangeId': { value: 0, type: SettingItemType.Int, public: false },
            'revisionService.lastProcessedChangeId': { value: 0, type: SettingItemType.Int, public: false },
            'searchEngine.initialIndexingDone': { value: false, type: SettingItemType.Bool, public: false },
            'searchEngine.lastProcessedResource': { value: '', type: SettingItemType.String, public: false },
            'revisionService.enabled': { section: 'revisionService', storage: SettingStorage.File, value: true, type: SettingItemType.Bool, public: true, label: () => (0, locale_1._)('Enable note history') },
            'revisionService.ttlDays': {
                section: 'revisionService',
                value: 90,
                type: SettingItemType.Int,
                public: true,
                minimum: 1,
                maximum: 365 * 2,
                step: 1,
                unitLabel: (value = null) => {
                    return value === null ? (0, locale_1._)('days') : (0, locale_1._)('%d days', value);
                },
                label: () => (0, locale_1._)('Keep note history for'),
                storage: SettingStorage.File,
            },
            'revisionService.intervalBetweenRevisions': { section: 'revisionService', value: 1000 * 60 * 10, type: SettingItemType.Int, public: false },
            'revisionService.oldNoteInterval': { section: 'revisionService', value: 1000 * 60 * 60 * 24 * 7, type: SettingItemType.Int, public: false },
            'welcome.wasBuilt': { value: false, type: SettingItemType.Bool, public: false },
            'welcome.enabled': { value: true, type: SettingItemType.Bool, public: false },
            'camera.type': { value: 0, type: SettingItemType.Int, public: false, appTypes: [AppType.Mobile] },
            'camera.ratio': { value: '4:3', type: SettingItemType.String, public: false, appTypes: [AppType.Mobile] },
            'spellChecker.enabled': { value: true, type: SettingItemType.Bool, isGlobal: true, storage: SettingStorage.File, public: false },
            'spellChecker.language': { value: '', type: SettingItemType.String, isGlobal: true, storage: SettingStorage.File, public: false },
            'spellChecker.languages': { value: [], type: SettingItemType.Array, isGlobal: true, storage: SettingStorage.File, public: false },
            windowContentZoomFactor: {
                value: 100,
                type: SettingItemType.Int,
                public: false,
                appTypes: [AppType.Desktop],
                minimum: 30,
                maximum: 300,
                step: 10,
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'layout.folderList.factor': {
                value: 1,
                type: SettingItemType.Int,
                section: 'appearance',
                public: true,
                appTypes: [AppType.Cli],
                label: () => (0, locale_1._)('Notebook list growth factor'),
                description: () => (0, locale_1._)('The factor property sets how the item will grow or shrink ' +
                    'to fit the available space in its container with respect to the other items. ' +
                    'Thus an item with a factor of 2 will take twice as much space as an item with a factor of 1.' +
                    'Restart app to see changes.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'layout.noteList.factor': {
                value: 1,
                type: SettingItemType.Int,
                section: 'appearance',
                public: true,
                appTypes: [AppType.Cli],
                label: () => (0, locale_1._)('Note list growth factor'),
                description: () => (0, locale_1._)('The factor property sets how the item will grow or shrink ' +
                    'to fit the available space in its container with respect to the other items. ' +
                    'Thus an item with a factor of 2 will take twice as much space as an item with a factor of 1.' +
                    'Restart app to see changes.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'layout.note.factor': {
                value: 2,
                type: SettingItemType.Int,
                section: 'appearance',
                public: true,
                appTypes: [AppType.Cli],
                label: () => (0, locale_1._)('Note area growth factor'),
                description: () => (0, locale_1._)('The factor property sets how the item will grow or shrink ' +
                    'to fit the available space in its container with respect to the other items. ' +
                    'Thus an item with a factor of 2 will take twice as much space as an item with a factor of 1.' +
                    'Restart app to see changes.'),
                storage: SettingStorage.File,
                isGlobal: true,
            },
            'syncInfoCache': {
                value: '',
                type: SettingItemType.String,
                public: false,
            },
            isSafeMode: {
                value: false,
                type: SettingItemType.Bool,
                public: false,
                appTypes: [AppType.Desktop],
                storage: SettingStorage.Database,
            },
            lastSettingDefaultMigration: {
                value: -1,
                type: SettingItemType.Int,
                public: false,
            },
            wasClosedSuccessfully: {
                value: true,
                type: SettingItemType.Bool,
                public: false,
            },
            installedDefaultPlugins: {
                value: [],
                type: SettingItemType.Array,
                public: false,
                storage: SettingStorage.Database,
            },
            // The biometrics feature is disabled by default and marked as beta
            // because it seems to cause a freeze or slow down startup on
            // certain devices. May be the reason for:
            //
            // - https://discourse.joplinapp.org/t/on-android-when-joplin-gets-started-offline/29951/1
            // - https://github.com/laurent22/joplin/issues/7956
            'security.biometricsEnabled': {
                value: false,
                type: SettingItemType.Bool,
                label: () => `${(0, locale_1._)('Use biometrics to secure access to the app')} (Beta)`,
                description: () => 'Important: This is a beta feature and it is not compatible with certain devices. If the app no longer starts after enabling this or is very slow to start, please uninstall and reinstall the app.',
                public: true,
                appTypes: [AppType.Mobile],
            },
            'security.biometricsSupportedSensors': {
                value: '',
                type: SettingItemType.String,
                public: false,
                appTypes: [AppType.Mobile],
            },
            'security.biometricsInitialPromptDone': {
                value: false,
                type: SettingItemType.Bool,
                public: false,
                appTypes: [AppType.Mobile],
            },
            // 'featureFlag.syncAccurateTimestamps': {
            // 	value: false,
            // 	type: SettingItemType.Bool,
            // 	public: false,
            // 	storage: SettingStorage.File,
            // },
            // 'featureFlag.syncMultiPut': {
            // 	value: false,
            // 	type: SettingItemType.Bool,
            // 	public: false,
            // 	storage: SettingStorage.File,
            // },
            'sync.allowUnsupportedProviders': {
                value: -1,
                type: SettingItemType.Int,
                public: false,
            },
            'sync.shareCache': {
                value: null,
                type: SettingItemType.String,
                public: false,
            },
            'voiceTypingBaseUrl': {
                value: '',
                type: SettingItemType.String,
                public: true,
                appTypes: [AppType.Mobile],
                description: () => (0, locale_1._)('Leave it blank to download the language files from the default website'),
                label: () => (0, locale_1._)('Voice typing language files (URL)'),
                section: 'note',
            },
        };
        this.metadata_ = Object.assign({}, this.buildInMetadata_);
        this.metadata_ = Object.assign(Object.assign({}, this.metadata_), this.customMetadata_);
        if (this.constants_.env === Env.Dev)
            this.validateMetadata(this.metadata_);
        return this.metadata_;
    }
    static validateMetadata(md) {
        for (const [k, v] of Object.entries(md)) {
            if (v.isGlobal && v.storage !== SettingStorage.File)
                throw new Error(`Setting "${k}" is global but storage is not "file"`);
        }
    }
    static isBuiltinKey(key) {
        return key in this.buildInMetadata_;
    }
    static customCssFilePath(filename) {
        return `${this.value('rootProfileDir')}/${filename}`;
    }
    static skipDefaultMigrations() {
        logger.info('Skipping all default migrations...');
        this.setValue('lastSettingDefaultMigration', defaultMigrations.length - 1);
    }
    static applyDefaultMigrations() {
        logger.info('Applying default migrations...');
        const lastSettingDefaultMigration = this.value('lastSettingDefaultMigration');
        for (let i = 0; i < defaultMigrations.length; i++) {
            if (i <= lastSettingDefaultMigration)
                continue;
            const migration = defaultMigrations[i];
            logger.info(`Applying default migration: ${migration.name}`);
            if (this.isSet(migration.name)) {
                logger.info('Skipping because value is already set');
                continue;
            }
            else {
                logger.info(`Applying previous default: ${migration.previousDefault}`);
                this.setValue(migration.name, migration.previousDefault);
            }
        }
        this.setValue('lastSettingDefaultMigration', defaultMigrations.length - 1);
    }
    static applyUserSettingMigration() {
        // Function to translate existing user settings to new setting.
        // eslint-disable-next-line github/array-foreach -- Old code before rule was applied
        userSettingMigration.forEach(userMigration => {
            if (!this.isSet(userMigration.newName) && this.isSet(userMigration.oldName)) {
                this.setValue(userMigration.newName, userMigration.transformValue(this.value(userMigration.oldName)));
                logger.info(`Migrating ${userMigration.oldName} to ${userMigration.newName}`);
            }
        });
    }
    static featureFlagKeys(appType) {
        const keys = this.keys(false, appType);
        return keys.filter(k => k.indexOf('featureFlag.') === 0);
    }
    static validateKey(key) {
        if (!key)
            throw new Error('Cannot register empty key');
        if (key.length > 128)
            throw new Error(`Key length cannot be longer than 128 characters: ${key}`);
        if (!key.match(/^[a-zA-Z0-9_\-.]+$/))
            throw new Error(`Key must only contain characters /a-zA-Z0-9_-./ : ${key}`);
    }
    static validateType(type) {
        if (!Number.isInteger(type))
            throw new Error(`Setting type is not an integer: ${type}`);
        if (type < 0)
            throw new Error(`Invalid setting type: ${type}`);
    }
    static registerSetting(key, metadataItem) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (metadataItem.isEnum && !metadataItem.options)
                    throw new Error('The `options` property is required for enum types');
                this.validateKey(key);
                this.validateType(metadataItem.type);
                this.customMetadata_[key] = Object.assign(Object.assign({}, metadataItem), { value: this.formatValue(metadataItem.type, metadataItem.value) });
                // Clear cache
                this.metadata_ = null;
                this.keys_ = null;
                // Reload the value from the database, if it was already present
                const valueRow = yield this.loadOne(key);
                if (valueRow) {
                    this.cache_.push({
                        key: key,
                        value: this.formatValue(key, valueRow.value),
                    });
                }
                this.dispatch({
                    type: 'SETTING_UPDATE_ONE',
                    key: key,
                    value: this.value(key),
                });
            }
            catch (error) {
                error.message = `Could not register setting "${key}": ${error.message}`;
                throw error;
            }
        });
    }
    static registerSection(name, source, section) {
        return __awaiter(this, void 0, void 0, function* () {
            this.customSections_[name] = Object.assign(Object.assign({}, section), { name: name, source: source });
        });
    }
    static settingMetadata(key) {
        const metadata = this.metadata();
        if (!(key in metadata))
            throw new JoplinError_1.default(`Unknown key: ${key}`, 'unknown_key');
        const output = Object.assign({}, metadata[key]);
        output.key = key;
        return output;
    }
    // Resets the key to its default value.
    static resetKey(key) {
        const md = this.settingMetadata(key);
        this.setValue(key, md.value);
    }
    static keyExists(key) {
        return key in this.metadata();
    }
    static isSet(key) {
        return !!this.cache_.find(d => d.key === key);
    }
    static keyDescription(key, appType = null) {
        const md = this.settingMetadata(key);
        if (!md.description)
            return null;
        return md.description(appType);
    }
    static isSecureKey(key) {
        return this.metadata()[key] && this.metadata()[key].secure === true;
    }
    static keys(publicOnly = false, appType = null, options = null) {
        options = Object.assign({ secureOnly: false }, options);
        if (!this.keys_) {
            const metadata = this.metadata();
            this.keys_ = [];
            for (const n in metadata) {
                if (!metadata.hasOwnProperty(n))
                    continue;
                this.keys_.push(n);
            }
        }
        if (appType || publicOnly || options.secureOnly) {
            const output = [];
            for (let i = 0; i < this.keys_.length; i++) {
                const md = this.settingMetadata(this.keys_[i]);
                if (publicOnly && !md.public)
                    continue;
                if (appType && md.appTypes && md.appTypes.indexOf(appType) < 0)
                    continue;
                if (options.secureOnly && !md.secure)
                    continue;
                output.push(md.key);
            }
            return output;
        }
        else {
            return this.keys_;
        }
    }
    static isPublic(key) {
        return this.keys(true).indexOf(key) >= 0;
    }
    // Low-level method to load a setting directly from the database. Should not be used in most cases.
    static loadOne(key) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.keyStorage(key) === SettingStorage.File) {
                let fileSettings = yield this.fileHandler.load();
                const md = this.settingMetadata(key);
                if (md.isGlobal) {
                    const rootFileSettings = yield this.rootFileHandler.load();
                    fileSettings = (0, mergeGlobalAndLocalSettings_1.default)(rootFileSettings, fileSettings);
                }
                return {
                    key,
                    value: fileSettings[key],
                };
            }
            // Always check in the database first, including for secure settings,
            // because that's where they would be if the keychain is not enabled (or
            // if writing to the keychain previously failed).
            //
            // https://github.com/laurent22/joplin/issues/5720
            const row = yield this.modelSelectOne('SELECT * FROM settings WHERE key = ?', [key]);
            if (row)
                return row;
            if (this.settingMetadata(key).secure) {
                return {
                    key,
                    value: yield this.keychainService().password(`setting.${key}`),
                };
            }
            return null;
        });
    }
    static load() {
        return __awaiter(this, void 0, void 0, function* () {
            this.cancelScheduleSave();
            this.cancelScheduleChangeEvent();
            this.cache_ = [];
            const rows = yield this.modelSelectAll('SELECT * FROM settings');
            this.cache_ = [];
            const pushItemsToCache = (items) => {
                for (let i = 0; i < items.length; i++) {
                    const c = items[i];
                    if (!this.keyExists(c.key))
                        continue;
                    c.value = this.formatValue(c.key, c.value);
                    c.value = this.filterValue(c.key, c.value);
                    this.cache_.push(c);
                }
            };
            // Keys in the database takes precedence over keys in the keychain because
            // they are more likely to be up to date (saving to keychain can fail, but
            // saving to database shouldn't). When the keychain works, the secure keys
            // are deleted from the database and transferred to the keychain in saveAll().
            const rowKeys = rows.map((r) => r.key);
            const secureKeys = this.keys(false, null, { secureOnly: true });
            const secureItems = [];
            for (const key of secureKeys) {
                if (rowKeys.includes(key))
                    continue;
                const password = yield this.keychainService().password(`setting.${key}`);
                if (password) {
                    secureItems.push({
                        key: key,
                        value: password,
                    });
                }
            }
            const itemsFromFile = [];
            if (this.canUseFileStorage()) {
                let fileSettings = yield this.fileHandler.load();
                if (this.value('isSubProfile')) {
                    const rootFileSettings = yield this.rootFileHandler.load();
                    fileSettings = (0, mergeGlobalAndLocalSettings_1.default)(rootFileSettings, fileSettings);
                }
                for (const k of Object.keys(fileSettings)) {
                    itemsFromFile.push({
                        key: k,
                        value: fileSettings[k],
                    });
                }
            }
            pushItemsToCache(rows);
            pushItemsToCache(secureItems);
            pushItemsToCache(itemsFromFile);
            this.dispatchUpdateAll();
        });
    }
    static canUseFileStorage() {
        return this.allowFileStorage && !shim_1.default.mobilePlatform();
    }
    static keyStorage(key) {
        if (!this.canUseFileStorage())
            return SettingStorage.Database;
        const md = this.settingMetadata(key);
        return md.storage || SettingStorage.Database;
    }
    static toPlainObject() {
        const keys = this.keys();
        const keyToValues = {};
        for (let i = 0; i < keys.length; i++) {
            keyToValues[keys[i]] = this.value(keys[i]);
        }
        return keyToValues;
    }
    static dispatchUpdateAll() {
        this.dispatch({
            type: 'SETTING_UPDATE_ALL',
            settings: this.toPlainObject(),
        });
    }
    static setConstant(key, value) {
        if (!(key in this.constants_))
            throw new Error(`Unknown constant key: ${key}`);
        this.constants_[key] = value;
    }
    static setValue(key, value) {
        if (!this.cache_)
            throw new Error('Settings have not been initialized!');
        value = this.formatValue(key, value);
        value = this.filterValue(key, value);
        for (let i = 0; i < this.cache_.length; i++) {
            const c = this.cache_[i];
            if (c.key === key) {
                const md = this.settingMetadata(key);
                if (md.isEnum === true) {
                    if (!this.isAllowedEnumOption(key, value)) {
                        throw new Error((0, locale_1._)('Invalid option value: "%s". Possible values are: %s.', value, this.enumOptionsDoc(key)));
                    }
                }
                if (c.value === value)
                    return;
                this.changedKeys_.push(key);
                // Don't log this to prevent sensitive info (passwords, auth tokens...) to end up in logs
                // logger.info('Setting: ' + key + ' = ' + c.value + ' => ' + value);
                if ('minimum' in md && value < md.minimum)
                    value = md.minimum;
                if ('maximum' in md && value > md.maximum)
                    value = md.maximum;
                c.value = value;
                this.dispatch({
                    type: 'SETTING_UPDATE_ONE',
                    key: key,
                    value: c.value,
                });
                this.scheduleSave();
                this.scheduleChangeEvent();
                return;
            }
        }
        this.cache_.push({
            key: key,
            value: this.formatValue(key, value),
        });
        this.dispatch({
            type: 'SETTING_UPDATE_ONE',
            key: key,
            value: this.formatValue(key, value),
        });
        this.changedKeys_.push(key);
        this.scheduleSave();
        this.scheduleChangeEvent();
    }
    static incValue(key, inc) {
        return this.setValue(key, this.value(key) + inc);
    }
    static toggle(key) {
        return this.setValue(key, !this.value(key));
    }
    // this method checks if the 'value' passed is present in the Setting "Array"
    // If yes, then it just returns 'true'. If its not present then, it will
    // update it and return 'false'
    static setArrayValue(settingName, value) {
        const settingValue = this.value(settingName);
        if (settingValue.includes(value))
            return true;
        settingValue.push(value);
        this.setValue(settingName, settingValue);
        return false;
    }
    static objectValue(settingKey, objectKey, defaultValue = null) {
        const o = this.value(settingKey);
        if (!o || !(objectKey in o))
            return defaultValue;
        return o[objectKey];
    }
    static setObjectValue(settingKey, objectKey, value) {
        let o = this.value(settingKey);
        if (typeof o !== 'object')
            o = {};
        o[objectKey] = value;
        this.setValue(settingKey, o);
    }
    static deleteObjectValue(settingKey, objectKey) {
        const o = this.value(settingKey);
        if (typeof o !== 'object')
            return;
        delete o[objectKey];
        this.setValue(settingKey, o);
    }
    static deleteKeychainPasswords() {
        return __awaiter(this, void 0, void 0, function* () {
            const secureKeys = this.keys(false, null, { secureOnly: true });
            for (const key of secureKeys) {
                yield this.keychainService().deletePassword(`setting.${key}`);
            }
        });
    }
    static enumOptionsToValueLabels(enumOptions, order, options = null) {
        options = Object.assign({ labelKey: 'label', valueKey: 'value' }, options);
        const output = [];
        for (const value of order) {
            output.push({
                [options.valueKey]: value,
                [options.labelKey]: enumOptions[value],
            });
        }
        for (const k in enumOptions) {
            if (!enumOptions.hasOwnProperty(k))
                continue;
            if (order.includes(k))
                continue;
            output.push({
                [options.valueKey]: k,
                [options.labelKey]: enumOptions[k],
            });
        }
        return output;
    }
    static valueToString(key, value) {
        const md = this.settingMetadata(key);
        value = this.formatValue(key, value);
        if (md.type === SettingItemType.Int)
            return value.toFixed(0);
        if (md.type === SettingItemType.Bool)
            return value ? '1' : '0';
        if (md.type === SettingItemType.Array)
            return value ? JSON.stringify(value) : '[]';
        if (md.type === SettingItemType.Object)
            return value ? JSON.stringify(value) : '{}';
        if (md.type === SettingItemType.String)
            return value ? `${value}` : '';
        throw new Error(`Unhandled value type: ${md.type}`);
    }
    static filterValue(key, value) {
        const md = this.settingMetadata(key);
        return md.filter ? md.filter(value) : value;
    }
    static formatValue(key, value) {
        const type = typeof key === 'string' ? this.settingMetadata(key).type : key;
        if (type === SettingItemType.Int)
            return !value ? 0 : Math.floor(Number(value));
        if (type === SettingItemType.Bool) {
            if (typeof value === 'string') {
                value = value.toLowerCase();
                if (value === 'true')
                    return true;
                if (value === 'false')
                    return false;
                value = Number(value);
            }
            return !!value;
        }
        if (type === SettingItemType.Array) {
            if (!value)
                return [];
            if (Array.isArray(value))
                return value;
            if (typeof value === 'string')
                return JSON.parse(value);
            return [];
        }
        if (type === SettingItemType.Object) {
            if (!value)
                return {};
            if (typeof value === 'object')
                return value;
            if (typeof value === 'string')
                return JSON.parse(value);
            return {};
        }
        if (type === SettingItemType.String) {
            if (!value)
                return '';
            return `${value}`;
        }
        throw new Error(`Unhandled value type: ${type}`);
    }
    static value(key) {
        // Need to copy arrays and objects since in setValue(), the old value and new one is compared
        // with strict equality and the value is updated only if changed. However if the caller acquire
        // an object and change a key, the objects will be detected as equal. By returning a copy
        // we avoid this problem.
        function copyIfNeeded(value) {
            if (value === null || value === undefined)
                return value;
            if (Array.isArray(value))
                return value.slice();
            if (typeof value === 'object')
                return Object.assign({}, value);
            return value;
        }
        if (key in this.constants_) {
            const v = this.constants_[key];
            const output = typeof v === 'function' ? v() : v;
            if (output === 'SET_ME')
                throw new Error(`SET_ME constant has not been set: ${key}`);
            return output;
        }
        if (!this.cache_)
            throw new Error('Settings have not been initialized!');
        for (let i = 0; i < this.cache_.length; i++) {
            if (this.cache_[i].key === key) {
                return copyIfNeeded(this.cache_[i].value);
            }
        }
        const md = this.settingMetadata(key);
        return copyIfNeeded(md.value);
    }
    // This function returns the default value if the setting key does not exist.
    static valueNoThrow(key, defaultValue) {
        if (!this.keyExists(key))
            return defaultValue;
        return this.value(key);
    }
    static isEnum(key) {
        const md = this.settingMetadata(key);
        return md.isEnum === true;
    }
    static enumOptionValues(key) {
        const options = this.enumOptions(key);
        const output = [];
        for (const n in options) {
            if (!options.hasOwnProperty(n))
                continue;
            output.push(n);
        }
        return output;
    }
    static enumOptionLabel(key, value) {
        const options = this.enumOptions(key);
        for (const n in options) {
            if (n === value)
                return options[n];
        }
        return '';
    }
    static enumOptions(key) {
        const metadata = this.metadata();
        if (!metadata[key])
            throw new JoplinError_1.default(`Unknown key: ${key}`, 'unknown_key');
        if (!metadata[key].options)
            throw new Error(`No options for: ${key}`);
        return metadata[key].options();
    }
    static enumOptionsDoc(key, templateString = null) {
        if (templateString === null)
            templateString = '%s: %s';
        const options = this.enumOptions(key);
        const output = [];
        for (const n in options) {
            if (!options.hasOwnProperty(n))
                continue;
            output.push(sprintf(templateString, n, options[n]));
        }
        return output.join(', ');
    }
    static isAllowedEnumOption(key, value) {
        const options = this.enumOptions(key);
        return !!options[value];
    }
    // For example, if settings is:
    // { sync.5.path: 'http://example', sync.5.username: 'testing' }
    // and baseKey is 'sync.5', the function will return
    // { path: 'http://example', username: 'testing' }
    static subValues(baseKey, settings, options = null) {
        const includeBaseKeyInName = !!options && !!options.includeBaseKeyInName;
        const output = {};
        for (const key in settings) {
            if (!settings.hasOwnProperty(key))
                continue;
            if (key.indexOf(baseKey) === 0) {
                const subKey = includeBaseKeyInName ? key : key.substr(baseKey.length + 1);
                output[subKey] = settings[key];
            }
        }
        return output;
    }
    static saveAll() {
        return __awaiter(this, void 0, void 0, function* () {
            if (Setting.autoSaveEnabled && !this.saveTimeoutId_)
                return Promise.resolve();
            logger.debug('Saving settings...');
            shim_1.default.clearTimeout(this.saveTimeoutId_);
            this.saveTimeoutId_ = null;
            const keys = this.keys();
            const valuesForFile = {};
            const queries = [];
            queries.push(`DELETE FROM settings WHERE key IN ("${keys.join('","')}")`);
            for (let i = 0; i < this.cache_.length; i++) {
                const s = Object.assign({}, this.cache_[i]);
                const valueAsString = this.valueToString(s.key, s.value);
                if (this.isSecureKey(s.key)) {
                    // We need to be careful here because there's a bug in the macOS keychain that can
                    // make it fail to save a password. https://github.com/desktop/desktop/issues/3263
                    // So we try to set it and if it fails, we set it on the database instead. This is not
                    // ideal because they won't be encrypted, but better than losing all the user's passwords.
                    // The passwords would be set again on the keychain once it starts working again (probably
                    // after the user switch their computer off and on again).
                    //
                    // Also we don't control what happens on the keychain - the values can be edited or deleted
                    // outside the application. For that reason, we rewrite it every time the values are saved,
                    // even if, internally, they haven't changed.
                    // As an optimisation, we check if the value exists on the keychain before writing it again.
                    try {
                        const passwordName = `setting.${s.key}`;
                        const currentValue = yield this.keychainService().password(passwordName);
                        if (currentValue !== valueAsString) {
                            const wasSet = yield this.keychainService().setPassword(passwordName, valueAsString);
                            if (wasSet)
                                continue;
                        }
                        else {
                            // The value is already in the keychain - so nothing to do
                            // Make sure to `continue` here otherwise it will save the password
                            // in clear text in the database.
                            continue;
                        }
                    }
                    catch (error) {
                        logger.error(`Could not set setting on the keychain. Will be saved to database instead: ${s.key}:`, error);
                    }
                }
                if (this.keyStorage(s.key) === SettingStorage.File) {
                    valuesForFile[s.key] = s.value;
                }
                else {
                    queries.push(database_1.default.insertQuery(this.tableName(), {
                        key: s.key,
                        value: valueAsString,
                    }));
                }
            }
            yield BaseModel_1.default.db().transactionExecBatch(queries);
            if (this.canUseFileStorage()) {
                if (this.value('isSubProfile')) {
                    const { globalSettings, localSettings } = (0, splitGlobalAndLocalSettings_1.default)(valuesForFile);
                    const currentGlobalSettings = yield this.rootFileHandler.load();
                    // When saving to the root setting file, we preserve the
                    // existing settings, which are specific to the root profile,
                    // and add the global settings.
                    yield this.rootFileHandler.save(Object.assign(Object.assign({}, currentGlobalSettings), globalSettings));
                    yield this.fileHandler.save(localSettings);
                }
                else {
                    yield this.fileHandler.save(valuesForFile);
                }
            }
            logger.debug('Settings have been saved.');
        });
    }
    static scheduleChangeEvent() {
        if (this.changeEventTimeoutId_)
            shim_1.default.clearTimeout(this.changeEventTimeoutId_);
        this.changeEventTimeoutId_ = shim_1.default.setTimeout(() => {
            this.emitScheduledChangeEvent();
        }, 1000);
    }
    static cancelScheduleChangeEvent() {
        if (this.changeEventTimeoutId_)
            shim_1.default.clearTimeout(this.changeEventTimeoutId_);
        this.changeEventTimeoutId_ = null;
    }
    static emitScheduledChangeEvent() {
        if (!this.changeEventTimeoutId_)
            return;
        shim_1.default.clearTimeout(this.changeEventTimeoutId_);
        this.changeEventTimeoutId_ = null;
        if (!this.changedKeys_.length) {
            // Sanity check - shouldn't happen
            logger.warn('Trying to dispatch a change event without any changed keys');
            return;
        }
        const keys = this.changedKeys_.slice();
        this.changedKeys_ = [];
        eventManager_1.default.emit(eventManager_1.EventName.SettingsChange, { keys });
    }
    static scheduleSave() {
        if (!Setting.autoSaveEnabled)
            return;
        if (this.saveTimeoutId_)
            shim_1.default.clearTimeout(this.saveTimeoutId_);
        this.saveTimeoutId_ = shim_1.default.setTimeout(() => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.saveAll();
            }
            catch (error) {
                logger.error('Could not save settings', error);
            }
        }), 500);
    }
    static cancelScheduleSave() {
        if (this.saveTimeoutId_)
            shim_1.default.clearTimeout(this.saveTimeoutId_);
        this.saveTimeoutId_ = null;
    }
    static publicSettings(appType) {
        if (!appType)
            throw new Error('appType is required');
        const metadata = this.metadata();
        const output = {};
        for (const key in metadata) {
            if (!metadata.hasOwnProperty(key))
                continue;
            const s = Object.assign({}, metadata[key]);
            if (!s.public)
                continue;
            if (s.appTypes && s.appTypes.indexOf(appType) < 0)
                continue;
            s.value = this.value(key);
            output[key] = s;
        }
        return output;
    }
    static typeToString(typeId) {
        if (typeId === SettingItemType.Int)
            return 'int';
        if (typeId === SettingItemType.String)
            return 'string';
        if (typeId === SettingItemType.Bool)
            return 'bool';
        if (typeId === SettingItemType.Array)
            return 'array';
        if (typeId === SettingItemType.Object)
            return 'object';
        throw new Error(`Invalid type ID: ${typeId}`);
    }
    static sectionOrder() {
        return [
            'general',
            'application',
            'appearance',
            'sync',
            'encryption',
            'joplinCloud',
            'plugins',
            'markdownPlugins',
            'note',
            'revisionService',
            'server',
            'keymap',
            'tools',
            'export',
            'moreInfo',
        ];
    }
    static sectionSource(sectionName) {
        if (this.customSections_[sectionName])
            return this.customSections_[sectionName].source || SettingSectionSource.Default;
        return SettingSectionSource.Default;
    }
    static isSubSection(sectionName) {
        return ['encryption', 'application', 'appearance', 'joplinCloud'].includes(sectionName);
    }
    static groupMetadatasBySections(metadatas) {
        const sections = [];
        const generalSection = { name: 'general', metadatas: [] };
        const nameToSections = {};
        nameToSections['general'] = generalSection;
        sections.push(generalSection);
        for (let i = 0; i < metadatas.length; i++) {
            const md = metadatas[i];
            if (!md.section) {
                generalSection.metadatas.push(md);
            }
            else {
                if (!nameToSections[md.section]) {
                    nameToSections[md.section] = {
                        name: md.section,
                        metadatas: [],
                        source: this.sectionSource(md.section),
                    };
                    sections.push(nameToSections[md.section]);
                }
                nameToSections[md.section].metadatas.push(md);
            }
        }
        // for (const name in this.customSections_) {
        // 	nameToSections[name] = {
        // 		name: name,
        // 		source: this.customSections_[name].source,
        // 		metadatas: [],
        // 	};
        // }
        return sections;
    }
    static sectionNameToLabel(name) {
        if (name === 'general')
            return (0, locale_1._)('General');
        if (name === 'sync')
            return (0, locale_1._)('Synchronisation');
        if (name === 'appearance')
            return (0, locale_1._)('Appearance');
        if (name === 'note')
            return (0, locale_1._)('Note');
        if (name === 'folder')
            return (0, locale_1._)('Notebook');
        if (name === 'markdownPlugins')
            return (0, locale_1._)('Markdown');
        if (name === 'plugins')
            return (0, locale_1._)('Plugins');
        if (name === 'application')
            return (0, locale_1._)('Application');
        if (name === 'revisionService')
            return (0, locale_1._)('Note History');
        if (name === 'encryption')
            return (0, locale_1._)('Encryption');
        if (name === 'server')
            return (0, locale_1._)('Web Clipper');
        if (name === 'keymap')
            return (0, locale_1._)('Keyboard Shortcuts');
        if (name === 'joplinCloud')
            return (0, locale_1._)('Joplin Cloud');
        if (name === 'tools')
            return (0, locale_1._)('Tools');
        if (name === 'export')
            return (0, locale_1._)('Export');
        if (name === 'moreInfo')
            return (0, locale_1._)('More information');
        if (this.customSections_[name] && this.customSections_[name].label)
            return this.customSections_[name].label;
        return name;
    }
    static sectionDescription(name) {
        if (name === 'markdownPlugins')
            return (0, locale_1._)('These plugins enhance the Markdown renderer with additional features. Please note that, while these features might be useful, they are not standard Markdown and thus most of them will only work in Joplin. Additionally, some of them are *incompatible* with the WYSIWYG editor. If you open a note that uses one of these plugins in that editor, you will lose the plugin formatting. It is indicated below which plugins are compatible or not with the WYSIWYG editor.');
        if (name === 'general')
            return (0, locale_1._)('Notes and settings are stored in: %s', toSystemSlashes(this.value('profileDir'), process.platform));
        if (this.customSections_[name] && this.customSections_[name].description)
            return this.customSections_[name].description;
        return '';
    }
    static sectionMetadataToSummary(metadata) {
        var _a;
        // TODO: This is currently specific to the mobile app
        const sectionNameToSummary = {
            'general': (0, locale_1._)('Language, date format'),
            'appearance': (0, locale_1._)('Themes, editor font'),
            'sync': (0, locale_1._)('Sync, encryption, proxy'),
            'joplinCloud': (0, locale_1._)('Email To Note, login information'),
            'markdownPlugins': (0, locale_1._)('Media player, math, diagrams, table of contents'),
            'note': (0, locale_1._)('Geolocation, spellcheck, editor toolbar, image resize'),
            'revisionService': (0, locale_1._)('Toggle note history, keep notes for'),
            'tools': (0, locale_1._)('Logs, profiles, sync status'),
            'export': (0, locale_1._)('Export your data'),
            'moreInfo': (0, locale_1._)('Donate, website'),
        };
        return (_a = sectionNameToSummary[metadata.name]) !== null && _a !== void 0 ? _a : '';
    }
    static sectionNameToIcon(name, appType) {
        const nameToIconMap = {
            'general': 'icon-general',
            'sync': 'icon-sync',
            'appearance': 'icon-appearance',
            'note': 'icon-note',
            'folder': 'icon-notebooks',
            'plugins': 'icon-plugins',
            'markdownPlugins': 'fab fa-markdown',
            'application': 'icon-application',
            'revisionService': 'icon-note-history',
            'encryption': 'icon-encryption',
            'server': 'far fa-hand-scissors',
            'keymap': 'fa fa-keyboard',
            'joplinCloud': 'fa fa-cloud',
            'tools': 'fa fa-toolbox',
            'export': 'fa fa-file-export',
            'moreInfo': 'fa fa-info-circle',
        };
        // Icomoon icons are currently not present in the mobile app -- we override these
        // below.
        //
        // These icons come from react-native-vector-icons.
        // See https://oblador.github.io/react-native-vector-icons/
        const mobileNameToIconMap = {
            'general': 'fa fa-sliders-h',
            'sync': 'fa fa-sync',
            'appearance': 'fa fa-ruler',
            'note': 'fa fa-sticky-note',
            'revisionService': 'far fa-history',
            'plugins': 'fa fa-puzzle-piece',
            'application': 'fa fa-cog',
            'encryption': 'fa fa-key',
        };
        // Overridden?
        if (appType === AppType.Mobile && name in mobileNameToIconMap) {
            return mobileNameToIconMap[name];
        }
        if (name in nameToIconMap) {
            return nameToIconMap[name];
        }
        if (this.customSections_[name] && this.customSections_[name].iconName)
            return this.customSections_[name].iconName;
        return 'fas fa-cog';
    }
    static appTypeToLabel(name) {
        // Not translated for now because only used on Welcome notes (which are not translated)
        if (name === 'cli')
            return 'CLI';
        return name[0].toUpperCase() + name.substr(1).toLowerCase();
    }
}
Setting.schemaUrl = 'https://joplinapp.org/schema/settings.json';
// For backward compatibility
Setting.TYPE_INT = SettingItemType.Int;
Setting.TYPE_STRING = SettingItemType.String;
Setting.TYPE_BOOL = SettingItemType.Bool;
Setting.TYPE_ARRAY = SettingItemType.Array;
Setting.TYPE_OBJECT = SettingItemType.Object;
Setting.TYPE_BUTTON = SettingItemType.Button;
Setting.THEME_LIGHT = 1;
Setting.THEME_DARK = 2;
Setting.THEME_OLED_DARK = 22;
Setting.THEME_SOLARIZED_LIGHT = 3;
Setting.THEME_SOLARIZED_DARK = 4;
Setting.THEME_DRACULA = 5;
Setting.THEME_NORD = 6;
Setting.THEME_ARITIM_DARK = 7;
Setting.FONT_DEFAULT = 0;
Setting.FONT_MENLO = 1;
Setting.FONT_COURIER_NEW = 2;
Setting.FONT_AVENIR = 3;
Setting.FONT_MONOSPACE = 4;
Setting.LAYOUT_ALL = 0;
Setting.LAYOUT_EDITOR_VIEWER = 1;
Setting.LAYOUT_EDITOR_SPLIT = 2;
Setting.LAYOUT_VIEWER_SPLIT = 3;
Setting.DATE_FORMAT_1 = 'DD/MM/YYYY';
Setting.DATE_FORMAT_2 = 'DD/MM/YY';
Setting.DATE_FORMAT_3 = 'MM/DD/YYYY';
Setting.DATE_FORMAT_4 = 'MM/DD/YY';
Setting.DATE_FORMAT_5 = 'YYYY-MM-DD';
Setting.DATE_FORMAT_6 = 'DD.MM.YYYY';
Setting.DATE_FORMAT_7 = 'YYYY.MM.DD';
Setting.DATE_FORMAT_8 = 'YYMMDD';
Setting.DATE_FORMAT_9 = 'YYYY/MM/DD';
Setting.TIME_FORMAT_1 = 'HH:mm';
Setting.TIME_FORMAT_2 = 'h:mm A';
Setting.TIME_FORMAT_3 = 'HH.mm';
Setting.SHOULD_REENCRYPT_NO = 0; // Data doesn't need to be re-encrypted
Setting.SHOULD_REENCRYPT_YES = 1; // Data should be re-encrypted
Setting.SHOULD_REENCRYPT_NOTIFIED = 2; // Data should be re-encrypted, and user has been notified
Setting.SYNC_UPGRADE_STATE_IDLE = 0; // Doesn't need to be upgraded
Setting.SYNC_UPGRADE_STATE_SHOULD_DO = 1; // Should be upgraded, but waiting for user to confirm
Setting.SYNC_UPGRADE_STATE_MUST_DO = 2; // Must be upgraded - on next restart, the upgrade will start
Setting.customCssFilenames = {
    JOPLIN_APP: 'userchrome.css',
    RENDERED_MARKDOWN: 'userstyle.css',
};
// Contains constants that are set by the application and
// cannot be modified by the user:
Setting.constants_ = {
    env: Env.Undefined,
    isDemo: false,
    appName: 'joplin',
    appId: 'SET_ME',
    appType: 'SET_ME',
    resourceDirName: '',
    resourceDir: '',
    profileDir: '',
    rootProfileDir: '',
    tempDir: '',
    pluginDataDir: '',
    cacheDir: '',
    pluginDir: '',
    homeDir: '',
    flagOpenDevTools: false,
    syncVersion: 3,
    startupDevPlugins: [],
    isSubProfile: false,
};
Setting.autoSaveEnabled = true;
Setting.allowFileStorage = true;
Setting.metadata_ = null;
Setting.keychainService_ = null;
Setting.keys_ = null;
Setting.cache_ = [];
Setting.saveTimeoutId_ = null;
Setting.changeEventTimeoutId_ = null;
Setting.customMetadata_ = {};
Setting.customSections_ = {};
Setting.changedKeys_ = [];
Setting.fileHandler_ = null;
Setting.rootFileHandler_ = null;
Setting.settingFilename_ = 'settings.json';
Setting.buildInMetadata_ = null;
exports.default = Setting;
//# sourceMappingURL=Setting.js.map