"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const tesseract_js_1 = require("tesseract.js");
const OcrDriverBase_1 = require("../OcrDriverBase");
const time_1 = require("@joplin/utils/time");
const shim_1 = require("../../../shim");
const Logger_1 = require("@joplin/utils/Logger");
const logger = Logger_1.default.create('OcrDriverTesseract');
let workerId_ = 1;
const formatTesseractBoundingBox = (boundingBox) => {
    return [boundingBox.x0, boundingBox.x1, boundingBox.y0, boundingBox.y1];
};
// Empirically, it seems anything below 70 is not usable. Between 70 and 75 it's
// hit and miss, but often it's good enough that we should keep the result.
// Above this is usually reliable.
const minConfidence = 70;
class OcrDriverTesseract extends OcrDriverBase_1.default {
    constructor(tesseract, workerPath = null, corePath = null) {
        super();
        this.tesseract_ = null;
        this.workerPath_ = null;
        this.corePath_ = null;
        this.workers_ = {};
        this.tesseract_ = tesseract;
        this.workerPath_ = workerPath;
        this.corePath_ = corePath;
    }
    acquireWorker(language) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.workers_[language])
                this.workers_[language] = [];
            const existingWorker = this.workers_[language].find(w => !w.busy);
            if (existingWorker) {
                existingWorker.busy = true;
                return existingWorker;
            }
            const createWorkerOptions = {
                workerBlobURL: false,
            };
            if (this.workerPath_)
                createWorkerOptions.workerPath = this.workerPath_;
            if (this.corePath_)
                createWorkerOptions.corePath = this.corePath_;
            const worker = yield this.tesseract_.createWorker(language, tesseract_js_1.OEM.LSTM_ONLY, createWorkerOptions);
            const output = {
                id: workerId_++,
                instance: worker,
                busy: true,
            };
            logger.info(`Created worker: ${output.id}`);
            this.workers_[language].push(output);
            return output;
        });
    }
    dispose() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const [language, workers] of Object.entries(this.workers_)) {
                for (const w of workers) {
                    yield w.instance.terminate();
                }
                this.workers_[language] = [];
            }
        });
    }
    terminateWorker(id) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const [, workers] of Object.entries(this.workers_)) {
                const idx = workers.findIndex(w => w.id === id);
                if (idx < 0)
                    continue;
                yield workers[idx].instance.terminate();
                workers.splice(idx, 1);
                break;
            }
        });
    }
    releaseWorker(worker) {
        return __awaiter(this, void 0, void 0, function* () {
            worker.busy = false;
        });
    }
    recognize(language, filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            // eslint-disable-next-line no-async-promise-executor -- can't think of any way to handle the timeout without using `new Promise`
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                const worker = yield this.acquireWorker(language);
                let hasTimedOut = false;
                const terminateTimeout_ = shim_1.default.setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                    yield this.terminateWorker(worker.id);
                    hasTimedOut = true;
                    reject(new Error(`Recognize operation timed out on: ${filePath}`));
                }), 10 * time_1.Minute);
                let result = null;
                try {
                    result = yield worker.instance.recognize(filePath, {}, {
                        text: false,
                        blocks: true,
                        hocr: false,
                        tsv: false,
                    });
                }
                catch (e) {
                    const error = typeof e === 'string' ? new Error(e) : e;
                    error.message = `Recognition failed on: ${filePath}: ${error.message}`;
                    if (!hasTimedOut)
                        reject(error);
                    return;
                }
                if (hasTimedOut)
                    return;
                shim_1.default.clearTimeout(terminateTimeout_);
                yield this.releaseWorker(worker);
                const goodParagraphs = [];
                let goodLines = [];
                for (const paragraph of result.data.paragraphs) {
                    const lines = [];
                    for (const line of paragraph.lines) {
                        // If the line confidence is above the threshold we keep the
                        // whole text. The confidence of individual words will vary and
                        // may be below the treshold, but there's a chance they will
                        // still be correct if the line as a whole is well recognised.
                        if (line.confidence < minConfidence)
                            continue;
                        const goodWords = line.words.map(w => {
                            const output = {
                                t: w.text,
                                bb: formatTesseractBoundingBox(w.bbox),
                            };
                            if (w.baseline && w.baseline.has_baseline)
                                output.bl = formatTesseractBoundingBox(w.baseline);
                            return output;
                        });
                        lines.push({
                            words: goodWords,
                        });
                    }
                    goodParagraphs.push({
                        text: lines.map(l => l.words.map(w => w.t).join(' ')).join('\n'),
                    });
                    goodLines = goodLines.concat(lines);
                }
                resolve({
                    // Note that Tesseract provides a `.text` property too, but it's the
                    // concatenation of all lines, even those with a low confidence
                    // score, so we recreate it here based on the good lines.
                    text: goodParagraphs.map(p => p.text).join('\n'),
                    lines: goodLines,
                });
            }));
        });
    }
}
exports.default = OcrDriverTesseract;
//# sourceMappingURL=OcrDriverTesseract.js.map