"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BaseModel_1 = require("../../BaseModel");
const Note_1 = require("../../models/Note");
const Resource_1 = require("../../models/Resource");
const shim_1 = require("../../shim");
const test_utils_1 = require("../../testing/test-utils");
const types_1 = require("../database/types");
const SearchEngine_1 = require("./SearchEngine");
const newSearchEngine = () => {
    const engine = new SearchEngine_1.default();
    engine.setDb((0, test_utils_1.db)());
    return engine;
};
describe('SearchEngine.resources', () => {
    beforeEach(() => __awaiter(void 0, void 0, void 0, function* () {
        global.console = require('console');
        yield (0, test_utils_1.setupDatabaseAndSynchronizer)(1);
        yield (0, test_utils_1.switchClient)(1);
    }));
    it('should index resources', () => __awaiter(void 0, void 0, void 0, function* () {
        const engine = newSearchEngine();
        yield Resource_1.default.save({
            id: '00000000000000000000000000000001',
            mime: 'image/jpeg',
            title: 'Bonjour ça va ?',
            ocr_status: types_1.ResourceOcrStatus.Done,
            ocr_text: 'héllô, hôw äre yoù ?',
        }, { isNew: true });
        yield engine.syncTables();
        const normalized = yield (0, test_utils_1.db)().selectAll('select * from items_normalized');
        expect(normalized[0].title).toBe('bonjour ca va ?');
        expect(normalized[0].body).toBe('hello, how are you ?');
    }));
    it('should return notes associated with indexed resources', (() => __awaiter(void 0, void 0, void 0, function* () {
        const note1 = yield Note_1.default.save({});
        yield Note_1.default.save({});
        yield shim_1.default.attachFileToNote(note1, `${test_utils_1.ocrSampleDir}/testocr.png`);
        const resource = (yield Resource_1.default.all())[0];
        yield (0, test_utils_1.resourceService)().indexNoteResources();
        const ocrService = (0, test_utils_1.newOcrService)();
        yield ocrService.processResources();
        const searchEngine = newSearchEngine();
        yield searchEngine.syncTables();
        const results = yield searchEngine.search('lazy fox');
        expect(results.length).toBe(1);
        expect(results[0].id).toBe(note1.id);
        expect(results[0].item_id).toBe(resource.id);
        expect(results[0].item_type).toBe(BaseModel_1.ModelType.Resource);
        yield ocrService.dispose();
    })));
    it('should delete normalized data when a resource is deleted', () => __awaiter(void 0, void 0, void 0, function* () {
        const engine = newSearchEngine();
        const resource = yield Resource_1.default.save({
            id: '00000000000000000000000000000001',
            mime: 'image/jpeg',
            title: 'hello',
            ocr_status: types_1.ResourceOcrStatus.Done,
            ocr_text: 'hi',
        }, { isNew: true });
        yield engine.syncTables();
        expect((yield (0, test_utils_1.db)().selectAll('select * from items_normalized')).length).toBe(1);
        yield Resource_1.default.delete(resource.id);
        expect((yield (0, test_utils_1.db)().selectAll('select * from items_normalized')).length).toBe(0);
    }));
    it('should sort resources', () => __awaiter(void 0, void 0, void 0, function* () {
        const engine = newSearchEngine();
        const resourceData = [
            ['abcd abcd abcd', 'efgh'],
            ['abcd', 'ijkl'],
            ['ijkl', 'mnop'],
        ];
        for (const [title, body] of resourceData) {
            yield Resource_1.default.save({
                mime: 'image/jpeg',
                title,
                ocr_status: types_1.ResourceOcrStatus.Done,
                ocr_text: body,
            });
            yield (0, test_utils_1.msleep)(1);
        }
        yield engine.syncTables();
        const results = yield engine.search('abcd', { includeOrphanedResources: true });
        expect(results[0].title).toBe('abcd abcd abcd');
        expect(results[1].title).toBe('abcd');
    }));
});
//# sourceMappingURL=SearchEngine.resources.test.js.map